from collections import Counter
import random

import numpy as np
from scipy.special import softmax
from torch import Tensor, FloatTensor


class BoostSampler:

    def __init__(self, data, batch_size=2**12):
        self.data = data
        self.user_num = len(data[1][0])
        self.user_proba = np.repeat(1/self.user_num, self.user_num)
        self.batch_size = batch_size
        self.user_idx = [x.item() for x in np.argmax(data[1], axis=1)]

        self.per_user_data = self.split_users()

    def get_epoch(self):
        """
        Supply elements based on the user_proba probabilities.
        """
        # There are two ways we can choose users:
        # 1. Choose every sample randomly, given a user distribution
        # 2. Take the amount of samples based on the distribution
        # Below, the first method is implemented.
        chosen = Counter([np.random.choice(range(len(self.user_proba)), p=self.user_proba) for _ in range(self.batch_size)])
        print(self.user_proba)
        print(chosen)
        # Then we have to choose data rows, based on the chosen indices
        data = []
        for user in np.unique(self.user_idx):
            ch = random.sample(self.per_user_data[user], chosen[user])
            data.extend(ch)
        return Tensor([x[0].tolist() for x in data]), [Tensor([x[1].tolist() for x in data]), Tensor([x[2].tolist() for x in data])]

    def update_user_proba(self, y_test, y_pred):
        y_test = y_test.tolist()
        y_pred = [x.item() for x in y_pred]

        hits = np.zeros(len(np.unique(y_test)))
        all = np.zeros(len(np.unique(y_test)))
        for i, real in enumerate(y_test):
            all[real] += 1
            if real == y_pred[i]:
                hits[real] += 1
        a = hits/all
        a = 1 - a
        self.user_proba = softmax(a)

    def split_users(self):
        """
        Split users into separate dataframes, so we can quickly sample them later on.
        """
        users = {}
        for user in np.unique(self.user_idx):
            idx = np.where(self.user_idx == user)[0]
            users[user] = [(self.data[0][i], self.data[1][i], self.data[2][i]) for i in idx]
        return users