import os
from datetime import datetime

import torch

import settings


class ModelHandler:

    def __init__(self, save_directory=settings.GARAGE, server=True):
        if server:
            self.save_directory = save_directory + datetime.now().strftime("%Y-%m-%d_%H:%M:%S") + "/"
            if not os.path.exists(self.save_directory):
                os.makedirs(self.save_directory)
        else:
            self.save_directory = save_directory + sorted(os.listdir(settings.GARAGE))[-1] + "/"

    def save_model(self, name, model, optimizer, epoch, utility_accuracy, privacy_accuracy):
        torch.save({
            "epoch": epoch,
            "model_state_dict": model.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "utility_accuracy": utility_accuracy,
            "privacy_accuracy": privacy_accuracy,
        }, self.save_directory + name)

    def reload_model(self, model_name, model, optimizer):
        """
        model and optimizer are initialized objects that fit the architecture of the parameters
        """
        checkpoint = torch.load(self.save_directory + model_name)
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        print("Model loaded, training progress:")
        print("Epoch: {}".format(checkpoint["epoch"]))
        print("Utility acc: {}".format(checkpoint["utility_accuracy"]))
        print("Privacy acc: {}".format(checkpoint["privacy_accuracy"]))
        print("Model set in training mode.")
        model.train()

        return model, optimizer