import io
import pickle
import time
from _socket import SHUT_RDWR
from datetime import datetime, timedelta
from socket import socket
from threading import Thread

import cloudpickle

import settings


class Buffer:
    def __init__(self, rhost, rport, shost, sport, name):
        self.rhost = rhost
        self.rport = rport
        self.shost = shost
        self.sport = sport
        self.name = name

        self.receive_socket = socket()
        self.receive_socket.bind((self.rhost, self.rport))

        self.data = []

        self.stop = False

        self.listen_thread = Thread(
            name=name, target=self.listen, args=(self.receive_socket,)
        )
        self.listen_thread.start()

    def listen(self, receive_socket):
        # Receive data until stop flag is set
        while not self.stop:
            receive_socket.listen()
            try:
                conn, addr = receive_socket.accept()
            except OSError:
                return
            actual_data = b""
            with conn:
                while not self.stop:
                    data = conn.recv(1048576)
                    if data is None or len(data) == 0:
                        break
                    actual_data += data
                self.data.append(cloudpickle.loads(actual_data))

    def take(self):
        timeout = datetime.now() + timedelta(seconds=settings.TIMEOUT)
        while len(self.data) < 1 and datetime.now() < timeout:
            time.sleep(1)
        if len(self.data) < 1:
            self.halt()
            raise TimeoutError("No data received.")
        return self.data.pop(0)

    def send(self, X):
        self.send_socket = socket()
        self.send_socket.connect((self.shost, self.sport))
        self.send_socket.send(cloudpickle.dumps(X))
        self.send_socket.close()

    def halt(self):
        self.stop = True
        self.receive_socket.shutdown(SHUT_RDWR)
        self.receive_socket.close()
        self.listen_thread.join()


class ServerBuffer(Buffer):
    def __init__(
        self,
        rhost=settings.SOCKET_SERVER_HOST,
        rport=settings.SOCKET_SERVER_PORT,
        shost=settings.SOCKET_CLIENT_HOST,
        sport=settings.SOCKET_CLIENT_PORT,
        name="Server",
    ):
        if name != "Server":
            name += " Server"
        super().__init__(rhost, rport, shost, sport, name)

    def take_forward(self):
        return self.take()

    def send_backward(self, X):
        self.send(X)


class ClientBuffer(Buffer):
    def __init__(
        self,
        rhost=settings.SOCKET_CLIENT_HOST,
        rport=settings.SOCKET_CLIENT_PORT,
        shost=settings.SOCKET_SERVER_HOST,
        sport=settings.SOCKET_SERVER_PORT,
        name="Client",
    ):
        if name != "Client":
            name += " Client"
        super().__init__(rhost, rport, shost, sport, name)

    def take_backward(self):
        return self.take()

    def send_forward(self, X):
        self.send(X)

    def take_predictions(self):
        timeout = datetime.now() + timedelta(seconds=settings.TIMEOUT)
        while len(self.data) < 3 and datetime.now() < timeout:
            time.sleep(1)
        if len(self.data) < 3:
            self.halt()
            raise TimeoutError("No data received.")
        return self.data.pop(2)
