from torch import nn, flatten, quantization

import settings


class ClientEncoderNetwork(nn.Module):
    def __init__(self):
        super(ClientEncoderNetwork, self).__init__()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(87, 72),
            nn.ReLU(),
        )

    def forward(self, x):
        return self.linear_relu_stack(x)


h1 = 512
l1 = h1 * 7

# This was used for first round of testing
# h1 = 64
# l1 = h1 * 7

# h1 = 32
# l1 = h1 * 7


class ClientEncoderLSTMNetwork(nn.Module):
    def __init__(self, input_size=87):
        super(ClientEncoderLSTMNetwork, self).__init__()
        self.lstm = nn.LSTM(input_size=input_size, hidden_size=h1)
        self.linear = nn.Linear(l1, h1)
        self.batch_norm = nn.BatchNorm1d(h1)
        self.dropout = nn.Dropout(settings.DROPOUT)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.lstm(x)
        x = flatten(x[0], start_dim=1)
        x = self.linear(x)
        x = self.dropout(x)
        x = self.batch_norm(x)
        x = self.relu(x)
        return x


class ClientEncoderGRUNetwork(nn.Module):
    def __init__(self, input_size=87):
        super(ClientEncoderGRUNetwork, self).__init__()

        if settings.QUANTIZATION:
            self.quant = quantization.QuantStub()

        self.lstm = nn.GRU(input_size=input_size, hidden_size=h1)
        self.linear = nn.Linear(l1, h1)
        self.batch_norm = nn.BatchNorm1d(h1)
        self.dropout = nn.Dropout(settings.DROPOUT)
        self.relu = nn.ReLU()

        if settings.QUANTIZATION:
            self.dequant = quantization.DeQuantStub()

    def forward(self, x):
        x = self.lstm(x)
        x = flatten(x[0], start_dim=1)

        if settings.QUANTIZATION:
            x = self.quant(x)

        x = self.linear(x)
        x = self.dropout(x)

        if not settings.QUANTIZATION:
            x = self.batch_norm(x)

        x = self.relu(x)

        if settings.QUANTIZATION:
            x = self.dequant(x)
        return x
