from torch import nn, cat

import settings


class ServerUtilityNetwork(nn.Module):
    def __init__(self):
        super(ServerUtilityNetwork, self).__init__()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(72, 64),
            nn.ReLU(),
            nn.Linear(64, 48),
            nn.ReLU(),
            nn.Linear(48, 36),
            nn.ReLU(),
            nn.Linear(36, 24),
            nn.ReLU(),
            nn.Linear(24, len(settings.CABACKEND_USERS)),
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        logits = self.linear_relu_stack(x)
        return logits


class ServerPrivacyNetwork(nn.Module):
    def __init__(self):
        super(ServerPrivacyNetwork, self).__init__()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(72, 64),
            nn.ReLU(),
            nn.Linear(64, 48),
            nn.ReLU(),
            nn.Linear(48, 36),
            nn.ReLU(),
            nn.Linear(36, 16),
            nn.ReLU(),
            nn.Linear(16, 3),
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        logits = self.linear_relu_stack(x)
        return logits


# split
# l1 = 256 * 3
# l2 = 128
# l3 = 64
# l4 = 32

# monolithic
# l1 = 256
# l2 = 128
# l3 = 64
# l4 = 32

# PESA
# l1 = 512
# l2 = 256
# l3 = 128
# l4 = 64

# Noise baseline
l1 = 609
l2 = 256
l3 = 128
l4 = 64


class ServerUtilityLSTMNetwork(nn.Module):
    def __init__(self):
        super(ServerUtilityLSTMNetwork, self).__init__()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(l1, l2),
            nn.BatchNorm1d(l2),
            nn.Dropout(settings.DROPOUT),
            nn.ReLU(),
            nn.Linear(l2, l3),
            nn.BatchNorm1d(l3),
            nn.Dropout(settings.DROPOUT),
            nn.ReLU(),
            nn.Linear(l3, l4),
            nn.BatchNorm1d(l4),
            nn.Dropout(settings.DROPOUT),
            nn.ReLU(),
            nn.Linear(l4, len(settings.CABACKEND_USERS)),
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        logits = self.linear_relu_stack(x)
        return logits


class ServerPrivacyLSTMNetwork(nn.Module):
    def __init__(self):
        super(ServerPrivacyLSTMNetwork, self).__init__()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(l1, l2),
            nn.BatchNorm1d(l2),
            nn.Dropout(settings.DROPOUT),
            nn.ReLU(),
            nn.Linear(l2, l3),
            nn.BatchNorm1d(l3),
            nn.Dropout(settings.DROPOUT),
            nn.ReLU(),
            nn.Linear(l3, l4),
            nn.BatchNorm1d(l4),
            nn.Dropout(settings.DROPOUT),
            nn.ReLU(),
            nn.Linear(l4, 3),
            nn.Softmax(dim=1),
        )

    def forward(self, x):
        logits = self.linear_relu_stack(x)
        return logits


class ServerUtilitySplitLSTMNetwork(nn.Module):
    def __init__(self):
        super(ServerUtilitySplitLSTMNetwork, self).__init__()
        self.lina = nn.Linear(l1 // 3, l1 // 3)
        self.linf = nn.Linear(l1 // 3, l1 // 3)
        self.linp = nn.Linear(l1 // 3, l1 // 3)

        self.lin1 = nn.Linear(l1, l2)
        self.bn1 = nn.BatchNorm1d(l2)
        self.dr1 = nn.Dropout(settings.DROPOUT)
        self.re1 = nn.ReLU()
        self.lin2 = nn.Linear(l2, l3)
        self.bn2 = nn.BatchNorm1d(l3)
        self.dr2 = nn.Dropout(settings.DROPOUT)
        self.re2 = nn.ReLU()
        self.lin3 = nn.Linear(l3, len(settings.CABACKEND_USERS))
        self.sm = nn.Softmax(dim=1)

    def forward(self, xa, xf, xp):
        # xa = self.lina(xa)
        # xf = self.lina(xf)
        # xp = self.lina(xp)
        x = cat((xa, xf, xp), dim=1)
        x = self.lin1(x)
        x = self.bn1(x)
        x = self.dr1(x)
        x = self.re1(x)
        x = self.lin2(x)
        x = self.bn2(x)
        x = self.dr2(x)
        x = self.re2(x)
        x = self.lin3(x)
        return self.sm(x)


class ServerPrivacySplitLSTMNetwork(nn.Module):
    def __init__(self):
        super(ServerPrivacySplitLSTMNetwork, self).__init__()
        self.lina = nn.Linear(l1 // 3, l1 // 3)
        self.linf = nn.Linear(l1 // 3, l1 // 3)
        self.linp = nn.Linear(l1 // 3, l1 // 3)

        self.lin1 = nn.Linear(l1, l2)
        self.bn1 = nn.BatchNorm1d(l2)
        self.dr1 = nn.Dropout(settings.DROPOUT)
        self.re1 = nn.ReLU()
        self.lin2 = nn.Linear(l2, l3)
        self.bn2 = nn.BatchNorm1d(l3)
        self.dr2 = nn.Dropout(settings.DROPOUT)
        self.re2 = nn.ReLU()
        self.lin3 = nn.Linear(l3, 3)
        self.sm = nn.Softmax(dim=1)

    def forward(self, xa, xf, xp):
        # xa = self.lina(xa)
        # xf = self.lina(xf)
        # xp = self.lina(xp)
        x = cat((xa, xf, xp), dim=1)
        x = self.lin1(x)
        x = self.bn1(x)
        x = self.dr1(x)
        x = self.re1(x)
        x = self.lin2(x)
        x = self.bn2(x)
        x = self.dr2(x)
        x = self.re2(x)
        x = self.lin3(x)
        return self.sm(x)
