from datetime import datetime
import os
from random import randint

import numpy as np
from keras.layers import BatchNormalization
from tensorflow.python.keras.losses import CategoricalCrossentropy, MeanSquaredError

import settings
from data.opportunity import Opportunity
from data.wesad import Wesad

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
import wandb
import tensorflow as tf
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    f1_score,
    precision_score,
    recall_score,
)
from tensorflow.python.keras import Input, Model
from tensorflow.python.keras.layers import (
    Dense,
    Dropout,
    LSTM,
    Conv1D,
    Conv2D,
    MaxPool1D,
    Flatten,
)
from tensorflow.python.keras.optimizer_v2.adam import Adam
from tensorflow.python.ops.math_ops import argmax, reduce_euclidean_norm
from tensorflow.python.ops.numpy_ops.np_arrays import convert_to_tensor
from torch.utils.data import DataLoader

from data.cabackend import Cabackend

# DROPOUT = 0.1
LEARNING_RATE = 0.0001
# LEARNING_RATE = 0.00001
# LEARNING_RATE = 1
# BATCH_SIZE = 8192
BATCH_SIZE = 2048
EPOCHS = 500
# EPOCHS = 100
EPCH_MULT = 3
DROPOUT = 0.3
# DROPOUT = 0.7

NEGATIVE_WEIGHT = -1
POSITIVE_WEIGHT = 0.1

# DATA_SEED = randint(1, 100000)
DATA_SEED = 55796

# How many training epochs before we move to cumulative training (k from PAN)
K = 10
# K = 1

# METHOD = "pan"
# METHOD = "ours"
# METHOD = "singlegood"
# METHOD = "singlebad"

# DATASET = "cabackend"
DATASET = "wesad"

# USERLIM = None
# USERLIM = 12
USERLIM = 15

ENC_LSTM = 128
ENC_DENSE = 128

CLASS_01 = 64
CLASS_02 = 32


class MisterStop:
    def __init__(self):
        self.privacy_accs = []

    def check_stop(self):
        print("Checking accs to stop...")
        if self.privacy_accs[-1] < np.mean(self.privacy_accs) + 0.01:
            print("Privacy classifier not progressing, stopping the training process.")
            return True
        print(
            "Last value ({}) higher than adjusted mean ({})".format(
                self.privacy_accs[-1], np.mean(self.privacy_accs) + 0.01
            )
        )
        print("Continuing...")
        return False

    def add_acc(self, privacy_acc):
        self.privacy_accs.append(privacy_acc)


class PlsWorkModelCNN:
    def __init__(self, trainset, validset, testset):
        self.trainset = trainset
        self.validset = validset
        self.testset = testset
        self.run_times = []

        self.i = 0

        # disable_eager_execution()
        if DATASET == "cabackend":
            encoder_in = Input(shape=(7, 87), name="enc_in")
        elif DATASET == "wesad":
            encoder_in = Input(shape=(7, 6), name="enc_in")
        lstm = LSTM(
            ENC_LSTM,
            # dropout=DROPOUT,
            name="enc_01",
        )(encoder_in)

        # cnn_01 = Conv1D(
        #     256, kernel_size=3, padding="same", data_format="channels_first"
        # )(encoder_in)
        # bn_01 = BatchNormalization()(cnn_01)
        # maxpool_01 = MaxPool1D(2, data_format="channels_first")(bn_01)
        #
        # cnn_02 = Conv1D(
        #     512, kernel_size=3, padding="same", data_format="channels_first"
        # )(maxpool_01)
        # bn_02 = BatchNormalization()(cnn_02)
        # maxpool_02 = MaxPool1D(2, data_format="channels_first")(bn_02)

        # cnn_03 = Conv1D(
        #     512, kernel_size=3, padding="same", data_format="channels_first"
        # )(maxpool_02)
        # bn_03 = BatchNormalization()(cnn_03)
        # maxpool_03 = MaxPool1D(2, data_format="channels_first")(bn_03)

        # cnn_04 = Conv1D(
        #     512, kernel_size=3, padding="same", data_format="channels_first"
        # )(maxpool_03)
        # bn_04 = BatchNormalization()(cnn_04)
        # maxpool_04 = MaxPool1D(2, data_format="channels_first")(bn_04)

        # flat_01 = Flatten()(maxpool_02)

        drp_out = Dense(ENC_DENSE, activation="relu", name="enc_out")(lstm)

        good_in = Input(shape=ENC_DENSE, name="good_in")
        good_01 = Dense(CLASS_01, activation="relu", name="good_01")(good_in)
        good_drp_01 = Dropout(DROPOUT)(good_01)
        good_02 = Dense(CLASS_02, activation="relu", name="good_02")(good_drp_01)
        good_drp_02 = Dropout(DROPOUT)(good_02)
        if DATASET == "cabackend":
            good_out = Dense(USERLIM, activation="softmax", name="good_classifier")(
                good_drp_02
            )
        elif DATASET == "wesad":
            good_out = Dense(USERLIM, activation="softmax", name="good_classifier")(
                good_drp_02
            )

        bad_in = Input(shape=ENC_DENSE, name="bad_in")
        bad_01 = Dense(CLASS_01, activation="relu", name="bad_01")(bad_in)
        bad_drp_01 = Dropout(DROPOUT)(bad_01)
        bad_02 = Dense(CLASS_02, activation="relu", name="bad_02")(bad_drp_01)
        bad_drp_02 = Dropout(DROPOUT)(bad_02)
        if DATASET == "cabackend":
            bad_out = Dense(3, activation="softmax", name="bad_classifier")(bad_drp_02)
        elif DATASET == "wesad":
            bad_out = Dense(4, activation="softmax", name="bad_classifier")(bad_drp_02)

        rec_in = Input(shape=ENC_DENSE, name="bad_in")
        rec_01 = Dense(CLASS_01, activation="relu", name="bad_01")(rec_in)
        rec_02 = Dense(CLASS_02, activation="relu", name="bad_02")(rec_01)
        if DATASET == "cabackend":
            rec_out = Dense(87, activation="softmax", name="bad_classifier")(rec_02)
        elif DATASET == "wesad":
            rec_out = Dense(6, activation="softmax", name="bad_classifier")(rec_02)

        self.optimizer = Adam(learning_rate=LEARNING_RATE)
        self.loss_fn = CategoricalCrossentropy()
        self.loss_rec = MeanSquaredError()

        self.model_encoder = Model(inputs=encoder_in, outputs=drp_out)
        self.model_good = Model(inputs=good_in, outputs=good_out)
        self.model_bad = Model(inputs=bad_in, outputs=bad_out)
        self.model_rec = Model(inputs=rec_in, outputs=rec_out)

        self.stop_check = None

        training_params = {
            "Learning Rate": LEARNING_RATE,
            "Batch Size": BATCH_SIZE,
            "Epochs": EPOCHS,
            "Dropout": DROPOUT,
            "Users": settings.CABACKEND_USERS,
            "Positive weight": POSITIVE_WEIGHT,
            "Negative weight": NEGATIVE_WEIGHT,
            "Data seed": DATA_SEED,
            "Encoder lstm size": ENC_LSTM,
            "Encoder dense size": ENC_DENSE,
            "Dense layer 1 size": CLASS_01,
            "Dense layer 2 size": CLASS_02,
            "Method": METHOD,
            "Dataset": DATASET,
            "Users #": USERLIM,
        }
        wandb.init(project="ppontt", entity="colosal", config=training_params)

    def train(self):
        self.model_encoder.summary()
        self.model_good.summary()
        self.model_bad.summary()
        # self.model_rec.summary()
        adv_epocs = EPOCHS * EPCH_MULT

        # Original PAN
        if METHOD == "pan":
            # Not utilised here, but an easy fix for missing mister stop in reporting.
            self.stop_check = MisterStop()
            self.pan_step()
            self.adversary_step(epochs=5000)
        # Our PAN
        elif METHOD == "ours":
            for i in range(8):
                print("Loop #{}".format(i + 1))
                self.normal_step()
                self.stop_check = MisterStop()
                self.adversary_step(epochs=adv_epocs)
                if self.stop_check.check_stop():
                    break
            self.adversary_step(epochs=5000)
        elif METHOD == "singlegood":
            self.single_step(good=True)
        elif METHOD == "singlebad":
            self.single_step(good=False)
        else:
            raise ValueError("Invalid method selected.")

        wandb.finish()

    def normal_step(self):
        print("NORMAL MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            # print("K: ", end="")
            # for k in range(K):
            # print("{}, ".format(k + 1), end="")
            s = datetime.now()

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_user = convert_to_tensor(y[0])
                y_task = convert_to_tensor(y[1])

                self.step_01(X, y_user)
                self.step_02(X, y_task)
                # The privacy reconstructor step
                # self.step_02_rec(X)

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_user = convert_to_tensor(y[0])
                y_task = convert_to_tensor(y[1])

                self.step_03(X, y_user, y_task)
            # print()

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == EPOCHS else False,
            )

    def single_step(self, good=True):
        if good:
            print("SINGLE GOOD MODE")
        else:
            print("SINGLE BAD MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            s = datetime.now()

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_user = convert_to_tensor(y[0])
                y_task = convert_to_tensor(y[1])

                if good:
                    self.step_01(X, y_user)
                else:
                    self.step_02_privacy(X, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == EPOCHS else False,
            )

    def pan_step(self):
        print("PAN MODE")
        # Train the network as specified, batch by batch
        for e in range(EPOCHS):
            self.i += 1
            s = datetime.now()
            for k in range(K):
                for X, y in self.trainset:
                    X = convert_to_tensor(X.numpy())
                    y_user = convert_to_tensor(y[0])
                    y_task = convert_to_tensor(y[1])

                    self.step_01(X, y_user)
                    self.step_02(X, y_task)
                    # The privacy reconstructor step
                    self.step_02_rec(X)

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_user = convert_to_tensor(y[0])
                y_task = convert_to_tensor(y[1])

                self.step_03(X, y_user, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == EPOCHS else False,
            )

    def adversary_step(self, epochs=EPOCHS):
        print("ADVERSARY MODE")
        # Only do this when retraining the adversary classifier
        for e in range(epochs):
            self.i += 1
            s = datetime.now()

            for X, y in self.trainset:
                X = convert_to_tensor(X.numpy())
                y_task = convert_to_tensor(y[1])

                self.step_02(X, y_task)

            epoch_time = datetime.now() - s
            if e % 100 == 0:
                print("Epoch ({}) run time: {}".format(self.i, epoch_time))
            self.report(
                epoch_time,
                True if e % 100 == 0 else False,
                True if e + 1 == epochs else False,
                stop=True,
            )

    def step_01(self, X, y_user):
        # Train the good classifier
        logits = self.model_encoder(X)
        with tf.GradientTape() as tape:
            logits = self.model_good(logits, training=True)
            loss_value = self.loss_fn(y_user, logits)

        grads = tape.gradient(loss_value, self.model_good.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_good.trainable_weights))

        # Train the encoder
        with tf.GradientTape() as tape:
            logits = self.model_good(self.model_encoder(X, training=True))
            loss_value = self.loss_fn(y_user, logits)

        grads = tape.gradient(loss_value, self.model_encoder.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

    def step_02(self, X, y_task):
        # Train the bad classifier
        logits = self.model_encoder(X)
        with tf.GradientTape() as tape:
            logits = self.model_bad(logits, training=True)
            loss_value = self.loss_fn(y_task, logits)

        grads = tape.gradient(loss_value, self.model_bad.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_bad.trainable_weights))

    def step_02_rec(self, X):
        # Train the rec classifier
        logits = self.model_encoder(X)
        with tf.GradientTape() as tape:
            logits = self.model_rec(logits, training=True)
            loss_value = self.loss_rec(X[:, -1, :], logits)

        grads = tape.gradient(loss_value, self.model_rec.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_rec.trainable_weights))

    def step_02_privacy(self, X, y_task):
        # This is used when bad classifier is the good classifier
        # Train the bad classifier
        logits = self.model_encoder(X)
        with tf.GradientTape() as tape:
            logits = self.model_bad(logits, training=True)
            loss_value = self.loss_fn(y_task, logits)

        grads = tape.gradient(loss_value, self.model_bad.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_bad.trainable_weights))

        # Train the encoder
        with tf.GradientTape() as tape:
            logits = self.model_bad(self.model_encoder(X, training=True))
            loss_value = self.loss_fn(y_task, logits)

        grads = tape.gradient(loss_value, self.model_encoder.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

    def step_03(self, X, y_user, y_task):
        # Train good and encoder classifier
        with tf.GradientTape(persistent=True) as tape:
            logits = self.model_encoder(X, training=True)
            gl = self.model_good(logits)
            loss_good = self.loss_fn(y_user, gl) * POSITIVE_WEIGHT
            bl = self.model_bad(logits)
            loss_bad = self.loss_fn(y_task, bl) * NEGATIVE_WEIGHT
            # logits = self.model_rec(self.model_encoder(X, training=True))
            # loss_rec = self.loss_rec(X[:, -1, :], logits) * NEGATIVE_WEIGHT
            # loss_value = loss_good + loss_bad + loss_rec
            # loss_value = loss_good + loss_bad

        # grads = tape.gradient(loss_value, self.model_good.trainable_weights)
        # self.optimizer.apply_gradients(zip(grads, self.model_good.trainable_weights))
        params_good = self.model_encoder.trainable_weights
        params_good.extend(self.model_good.trainable_weights)
        grads = tape.gradient(
            loss_good,
            params_good,
        )
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

        params_bad = self.model_encoder.trainable_weights
        params_bad.extend(self.model_bad.trainable_weights)
        grads = tape.gradient(loss_bad, params_bad)
        self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

        # params_rec = self.model_encoder.trainable_weights
        # params_rec.extend(self.model_rec.trainable_weights)
        # grads = tape.gradient(loss_rec, params_rec)
        # self.optimizer.apply_gradients(zip(grads, self.model_encoder.trainable_weights))

    def report(self, epoch_time, cm=False, test=False, stop=False):
        self.run_times.append(epoch_time)

        evaluation = {}
        ul, tl, ua, ta, uf, tf, up, tp, ur, tr, ucm, tcm = self.evaluate(
            self.trainset, cm
        )
        evaluation["Utility/Loss/train"] = ul
        evaluation["Privacy/Loss/train"] = tl
        evaluation["Utility/Accuracy/train"] = ua
        evaluation["Privacy/Accuracy/train"] = ta
        evaluation["Utility/F1 Score/train"] = uf
        evaluation["Privacy/F1 Score/train"] = tf
        evaluation["Utility/Precision/train"] = up
        evaluation["Privacy/Precision/train"] = tp
        evaluation["Utility/Recall/train"] = ur
        evaluation["Privacy/Recall/train"] = tr
        if cm:
            evaluation["Utility/ConfusionMatrix/train"] = ucm
            evaluation["Privacy/ConfusionMatrix/train"] = tcm

        ul, tl, ua, ta, uf, tf, up, tp, ur, tr, ucm, tcm = self.evaluate(
            self.validset, cm
        )
        if stop:
            self.stop_check.add_acc(ta)
        evaluation["Utility/Loss/valid"] = ul
        evaluation["Privacy/Loss/valid"] = tl
        evaluation["Utility/Accuracy/valid"] = ua
        evaluation["Privacy/Accuracy/valid"] = ta
        evaluation["Utility/F1 Score/valid"] = uf
        evaluation["Privacy/F1 Score/valid"] = tf
        evaluation["Utility/Precision/valid"] = up
        evaluation["Privacy/Precision/valid"] = tp
        evaluation["Utility/Recall/valid"] = ur
        evaluation["Privacy/Recall/valid"] = tr
        if cm:
            evaluation["Utility/ConfusionMatrix/valid"] = ucm
            evaluation["Privacy/ConfusionMatrix/valid"] = tcm

        if test:
            ul, tl, ua, ta, uf, tf, up, tp, ur, tr, ucm, tcm = self.evaluate(
                self.testset, False
            )
            evaluation["Utility/Loss/test"] = ul
            evaluation["Privacy/Loss/test"] = tl
            evaluation["Utility/Accuracy/test"] = ua
            evaluation["Privacy/Accuracy/test"] = ta
            evaluation["Utility/F1 Score/test"] = uf
            evaluation["Privacy/F1 Score/test"] = tf
            evaluation["Utility/Precision/test"] = up
            evaluation["Privacy/Precision/test"] = tp
            evaluation["Utility/Recall/test"] = ur
            evaluation["Privacy/Recall/test"] = tr

            evaluation["Epoch time"] = np.mean(
                [x.total_seconds() for x in self.run_times]
            )

        wandb.log(evaluation)

    def evaluate(self, dataset, cm):
        user_accs = []
        task_accs = []
        user_loss = []
        task_loss = []
        user_f1 = []
        task_f1 = []
        user_prec = []
        task_prec = []
        user_reca = []
        task_reca = []
        if cm:
            trues, preds, ttrues, tpreds = [], [], [], []
        n = 0
        for X, y in dataset:
            X = convert_to_tensor(X.numpy())
            y_user = convert_to_tensor(y[0])
            y_task = convert_to_tensor(y[1])
            n += len(X)

            y_pred_user = self.model_good(self.model_encoder(X))
            y_pred_task = self.model_bad(self.model_encoder(X))
            user_loss.append(self.loss_fn(y_user, y_pred_user) * len(X))
            task_loss.append(self.loss_fn(y_task, y_pred_task) * len(X))

            y_user = argmax(y_user, 1)
            y_task = argmax(y_task, 1)
            y_pred_user = argmax(y_pred_user, 1)
            y_pred_task = argmax(y_pred_task, 1)
            user_accs.append(accuracy_score(y_user, y_pred_user) * len(X))
            task_accs.append(accuracy_score(y_task, y_pred_task) * len(X))
            user_f1.append(f1_score(y_user, y_pred_user, average="weighted") * len(X))
            task_f1.append(f1_score(y_task, y_pred_task, average="weighted") * len(X))
            user_prec.append(
                precision_score(
                    y_user, y_pred_user, average="weighted", zero_division=0.0
                )
                * len(X)
            )
            task_prec.append(
                precision_score(
                    y_task, y_pred_task, average="weighted", zero_division=0.0
                )
                * len(X)
            )
            user_reca.append(
                recall_score(y_user, y_pred_user, average="weighted") * len(X)
            )
            task_reca.append(
                recall_score(y_task, y_pred_task, average="weighted") * len(X)
            )
            # print(accuracy_score(y_user, y_pred_user))
            # print(f1_score(y_user, y_pred_user, average="micro"))
            # print(f1_score(y_user, y_pred_user, average="macro"))
            # print(f1_score(y_user, y_pred_user, average="weighted"))
            # print(f1_score(y_user, y_pred_user, average=None))
            # print()
            if cm:
                trues.extend(y_user)
                preds.extend(y_pred_user)
                ttrues.extend(y_task)
                tpreds.extend(y_pred_task)
        if cm:
            trues = [x.numpy() for x in trues]
            preds = [x.numpy() for x in preds]
            ttrues = [x.numpy() for x in ttrues]
            tpreds = [x.numpy() for x in tpreds]

            return (
                sum(user_loss) / n,
                sum(task_loss) / n,
                sum(user_accs) / n,
                sum(task_accs) / n,
                sum(user_f1) / n,
                sum(task_f1) / n,
                sum(user_prec) / n,
                sum(task_prec) / n,
                sum(user_reca) / n,
                sum(task_reca) / n,
                wandb.plot.confusion_matrix(
                    y_true=trues,
                    preds=preds,
                    # class_names=[str(x) for x in settings.CABACKEND_USERS],
                    class_names=[str(x) for x in range(15)],
                ),
                wandb.plot.confusion_matrix(
                    y_true=ttrues,
                    preds=tpreds,
                    class_names=["0", "1", "2", "3", "4", "5", "6", "7"],
                    # , class_names=["1", "2", "3"]
                ),
            )
        else:
            return (
                sum(user_loss) / n,
                sum(task_loss) / n,
                sum(user_accs) / n,
                sum(task_accs) / n,
                sum(user_f1) / n,
                sum(task_f1) / n,
                sum(user_prec) / n,
                sum(task_prec) / n,
                sum(user_reca) / n,
                sum(task_reca) / n,
                None,
                None,
            )


if __name__ == "__main__":
    for METHOD in ["singlegood", "singlebad", "pan"]:
        for DATASET in ["wesad", "cabackend"]:
            if DATASET == "cabackend":
                cabackend = Cabackend(
                    split=[0.70, 0.15, 0.15],
                    data_path="../data/cabackend/",
                    userlim=USERLIM,
                )
                train_data, validation_data, test_data = cabackend.get_lstm_data()
            elif DATASET == "wesad":
                wesad = Wesad(
                    data_dir="/home/andraz/Projects/ppontt_v2.1/data/wesad/",
                    seed=DATA_SEED,
                    userlim=USERLIM,
                )
                train_data, validation_data, test_data = wesad.get_lstm_data()
            else:
                raise ValueError("Unknown dataset selected.")

            train_loader = DataLoader(
                train_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=False
            )
            valid_loader = DataLoader(
                validation_data,
                batch_size=BATCH_SIZE,
                shuffle=False,
                drop_last=False,
            )
            test_loader = DataLoader(
                test_data, batch_size=BATCH_SIZE, shuffle=False, drop_last=False
            )

            # model = PlsWorkModel(train_loader, valid_loader, test_loader)
            model = PlsWorkModelCNN(train_loader, valid_loader, test_loader)
            model.train()
