from __future__ import division

import logging
import sys

import torch
import torch.backends.cudnn as cudnn

import settings

# from data.extrasensory import Extrasensory
from helpers.logger import setup_logger
from helpers.soc_comm import ServerBuffer
from models.server import (
    ServerPrivacyLSTMNetwork,
    ServerUtilityLSTMNetwork,
)
from training.server import train_end


def main():
    print("python version : {}".format(sys.version.replace("\n", " ")))
    print("torch  version : {}".format(torch.__version__))
    print("cudnn  version : {}".format(torch.backends.cudnn.version()))
    # device = "cpu"
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("device         : {}".format(device))
    print()

    # Init loggers
    setup_logger("Server", "log/server.log")
    server_logger = logging.getLogger("Server")

    # Init model, loss function, and optimizers
    # Encoder here is only initialised to write it in tensorboard
    utility_model = ServerUtilityLSTMNetwork().float().to(device)
    privacy_model = ServerPrivacyLSTMNetwork().float().to(device)

    loss = torch.nn.CrossEntropyLoss()

    utility_optimizer = torch.optim.Adam(
        utility_model.parameters(),
        lr=settings.LEARNING_RATE,
    )
    privacy_optimizer = torch.optim.Adam(
        privacy_model.parameters(),
        lr=settings.LEARNING_RATE,
    )

    # Socket communication
    buffer = ServerBuffer()
    # Run the training
    train_end(
        (utility_model, privacy_model),
        loss,
        (utility_optimizer, privacy_optimizer),
        buffer,
        server_logger,
        device,
    )


if __name__ == "__main__":
    main()
